{-# OPTIONS_GHC -Wall -XFlexibleInstances #-}
module ToINV (showINV, combineAnd, obtainQueues, obtainFlops, getINV, dispQOFN, QueueOrFlopNum(..), translateTerm, (.-.), (.+.), (.*.), LPMap, UntimedLPTerm, getPrimitives, echelonReduce, oneMap)  where
import Terms (Term(..))
import qualified Data.Set as Set
import qualified Data.IntSet as IntSet
import qualified Data.Map.Lazy as Map
import EParser (QueueAssignment(..),FlopAssignment(..),InputAssignment(..),OutputAssignment(..)
               ,OutputWire(..),HasName(name))
import Prelude hiding (init)
-- import Debug.Trace
type LPMap a = Map.Map a Int
type UntimedLPTerm = Set.Set String

(.*.) :: Int -> Map.Map a Int -> Map.Map a Int 
(.*.) x y = Map.map ((*) x) y
(.-.),(.+.) :: Ord a => Map.Map a Int -> Map.Map a Int -> Map.Map a Int
(.-.) = Map.mergeWithKey (\_ a b->case a-b of {0->Nothing;s->Just s}) id (Map.map (\x -> -x))
(.+.) = Map.mergeWithKey (\_ a b->case a+b of {0->Nothing;s->Just s}) id id

data TranslatableTerm
 = TR_OR    Term Term
 | TR_XOR   Term Term
 | TR_ITE   Term Term Term
 | TR_NOT   Term
 | TR_VALUE Bool
 | TR_RESET deriving Show

-- "translateTranslatableTerm" would be a nice candidate to prove correct in ACL2
translateTranslatableTerm :: [TranslatableTerm]
                          -> LPMap UntimedLPTerm -- the resulting LP variable
translateTranslatableTerm [] = (Map.singleton Set.empty 1)
translateTranslatableTerm (ini:bs)
 = combineAnd (case ini of
                (TR_RESET    ) -> (Map.empty)
                (TR_VALUE b  ) -> if b then (Map.singleton Set.empty 1)
                                       else (Map.empty)
                (TR_ITE a b c) -> translateTerm (T_AND a b) .+. translateTerm (T_AND (T_NOT a) c) -- note: this is more efficient than translating T_ITE to a T_OR: we know that both sides of the T_OR are mutually exclusive and use that here
                (TR_OR  y1 y2) -> let t1=translateTerm ( y1)
                                      t2=translateTerm ( y2)
                                  in -- (Map.singleton Set.empty 1) .-. (combineAnd t1 t2) -- 
                                     -- note: this was (on positive terms t1 t2)
                                      t1 .+. t2 .-. combineAnd t1 t2
                                     -- which may be slower by creating t1 and t2 as separate terms
                                     -- (it is just as fast if t1 and t2 are positive terms, however)
                (TR_XOR y1 y2) -> let t1=translateTerm y1
                                      t2=translateTerm y2
                                  in t1 .+. t2 .-. (2 .*. combineAnd t1 t2)
                (TR_NOT y1   ) -> oneMap .-. translateTerm y1
              )
              (translateTranslatableTerm bs)

oneMap :: LPMap UntimedLPTerm
oneMap = Map.singleton Set.empty 1

-- "combineAnd" would be a nice candidate to prove correct in ACL2
combineAnd :: LPMap UntimedLPTerm -> LPMap UntimedLPTerm -- two LPs for which the AND must be calculated
           -> LPMap UntimedLPTerm -- result
-- this is just some sort of Cartesian product
combineAnd a b = Map.filter ((/=) 0) $ 
                 Map.fromListWith (+) [(Set.union a' b', ai*bi) | (a',ai) <- Map.toList a, (b',bi) <- Map.toList b]

-- compute the LP-term for (x && y) in which x is a list of primitives and y is an LP-term
extendLPMap :: Set.Set String -> LPMap UntimedLPTerm -> LPMap UntimedLPTerm
extendLPMap s v = Map.filter ((/=) 0) $ 
                  Map.mapKeysWith (+) (Set.union s) v

translateTerm :: Term -> LPMap UntimedLPTerm
translateTerm t = extendLPMap (Set.fromList s) (translateTranslatableTerm tt)
 where (s,tt) = collectAnds t
       collectAnds :: Term -> ([String],[TranslatableTerm])
       -- collectAnds:
       -- returns Either:
       --   Left: a set of strings which is to be used for declarations
       --   Right: tuple: ( list of conjuncts that are primitive and succesfully translated as a set of strings to be used for declarations
       --                 , list of translatables (i.e. not T_AND or primitive)
       --                   which may still contain T_AND in a nested place (such as T_NOT (T_AND ...))
       --                 )
       collectAnds (T_QTRDY x  ) = trm  "QTRDY" x
       collectAnds (T_QIRDY x  ) = trm  "QIRDY" x
       collectAnds (T_QDATA x y) = trm' "QDATA" x y
       collectAnds (T_OTRDY x  ) = trm  "OTRDY" x
       collectAnds (T_IIRDY x  ) = trm  "IIRDY" x
       collectAnds (T_IDATA x y) = trm' "IDATA" x y
       collectAnds (T_FLOPV x  ) = trm  "FLOPV" x
       collectAnds (T_UNKNOWN x) = trm  "UNKNO" (show x)
       collectAnds (T_INPUT x  ) = trm  "INPUT" x
       collectAnds (T_AND x y)
         = case (collectAnds x,collectAnds y) of
             ~((nx,ox),(ny,oy)) -> (nx++ny, ox++oy)
       collectAnds (T_OR  x y)   = wrong (TR_OR    x y)
       collectAnds (T_XOR x y)   = wrong (TR_XOR   x y)
       collectAnds (T_ITE x y z) = wrong (TR_ITE   x y z)
       collectAnds (T_NOT x)     = wrong (TR_NOT   x)
       collectAnds (T_VALUE x)   = wrong (TR_VALUE x)
       collectAnds (T_RESET)     = wrong (TR_RESET)
       trm x y = ([x++y],[])
       trm' x y u = ([x++y++"-"++show u],[])
       wrong x = ([],[x])

tnot :: Term -> Term
tnot (T_NOT a) = a
tnot (T_XOR a b) = (T_XOR (tnot a) b)
tnot (T_OR a b) = (T_AND (tnot a) (tnot b))
tnot b = (T_NOT b)

getPrimitives :: Ord a => [Map.Map a a1] -> Set.Set a
getPrimitives lst = Set.unions (map Map.keysSet lst)

obtainDataWires :: QueueAssignment -> [(String, (QueueAssignment, Int))]
obtainDataWires o@(QueueAssignment qname _size _inIrdy idata _ouTrdy)
 = [(head (Set.toList (head (Map.keys (translateTerm (T_QDATA qname nr))))), (o,nr)) | nr<-[0..(length idata - 1)]]

showINV :: Bool -- obtain enter exit
        -> Bool -- obtain flops?
        -> ([QueueAssignment],[FlopAssignment],[InputAssignment],[OutputAssignment],[OutputWire]) -> IO ()
showINV getInOut getFlops (queues,flops,_ins,_outs,_wires)
 = putStrLn$ showEQS (getINV getInOut queues (if getFlops then flops else []))
 where showEQS [] = []
       showEQS ((negs, poss):lst) = (showAdd negs ++ " = " ++ showAdd poss) ++ "\n" ++ rst
        where rst = showEQS lst
       showT (v,nm) = if (v == 1) then (dispQOFN nm) else (show v) ++ " * " ++ (dispQOFN nm)
       showAdd [] = "0" 
       showAdd [e] = showT e
       showAdd (e:bs) = showT e ++ " + " ++ showAdd bs

getINV :: Bool -> [QueueAssignment]
               -> [FlopAssignment]
               -> [([(Int, QueueOrFlopNum)], [(Int, QueueOrFlopNum)])]
getINV getInOut queues flops
 = genEQS sol
 where preQueueInfo   = obtainQueues getInOut queues ++ obtainFlops getInOut flops ++ [(N, translateTerm (T_VALUE True))]
       primitives     = Set.toList (getPrimitives (map snd queueInfo))
       dataWires      = Map.fromList$ concat (map obtainDataWires queues)
       queueInfo      = preQueueInfo ++ fixedPointExtraQueue getInOut dataWires (getPrimitives (map snd preQueueInfo))
       nrOfPrims      = length primitives
       primNr = Map.fromList (zip primitives [0..])
       sol = map Map.toList (echelonReduce [ Map.fromList ([(Map.findWithDefault (error "not found!") p primNr,v) | (p,v)<-Map.toList qi, v/=0]
                                                   ++ [(j,1)])
                                  | (j,(_,qi))<- zip [(length primitives)..] queueInfo])
       genEQS [] = []
       genEQS ([]:lst) = genEQS lst
       genEQS (a@((f,_):_):lst) = if f < nrOfPrims then rst else (negs, poss) : rst
         where poss = [( (v), fst (queueInfo !! (j-nrOfPrims))) | (j,v)<-a, v> 0]
               negs = [(-(v), fst (queueInfo !! (j-nrOfPrims))) | (j,v)<-a, v< 0]
               rst = genEQS lst

echelonReduce :: Ord k => [Map.Map k Int] -> [Map.Map k Int]
echelonReduce (a:rst) = if (Map.null a) then echelonReduce rst else a:echelonReduce (map sweep rst)
  where (ll,ht) = Map.findMin a
        sweep v = case Map.lookup ll v of
                    Nothing -> v
                    Just f' -> if f' == ht then v .-. a -- even though normalization might be possible, it is not (yet) required here. This reduces runtime with about 25% in practice
                               else if f'==(-ht) then v .+. a -- same here
                               else normalize (Map.map ((*) (-ht)) v .+. Map.map ((*) f') a)
echelonReduce [] = []
normalize :: Integral b => Map.Map k b -> Map.Map k b
normalize sp = Map.map (\f -> quot f factor) sp
 where factor = Map.foldr gcd 0 sp

queueName :: QueueAssignment -> String
queueName (QueueAssignment qname _ _ _ _) = qname

fixedPointExtraQueue :: Bool -> Map.Map String (QueueAssignment, Int) -> Set.Set (Set.Set String) -> [(QueueOrFlopNum, Map.Map UntimedLPTerm Int)]
fixedPointExtraQueue ee datas prims
 = nub' Set.empty (getPrimSet (Set.toList prims))
 where
  getPrimSet ps
   = concat 
     [ Map.toList $
           Map.fromListWith (\(a,x) (_,y) -> (a,x++y)) 
                            [ (queueName q,(q,[nr]))
                            | var <- Set.toList prim
                            , Just (q, nr) <- [Map.lookup var datas]
                            ]
          | prim <- ps]
  nub' _ [] = []
  nub' prev ((qnm,(q,nrs)):rst)
   = if (Set.member (qnm,nrs) prev) then nub' prev rst else
     (if ee then (QNB q (IntSet.fromList nrs) Up, res1)
                 :(QNB q (IntSet.fromList nrs) Down, res2)
                 :nubVal
             else (QNB q (IntSet.fromList nrs) Val, res):nubVal)
   where (res1,res2) = (obtainExtraQueue q nrs)
         res = res1 .-. res2
         nubVal = nub' (Set.insert (qnm,nrs) prev) (rst ++ [ps | ps@(a,(_,c)) <- getPrimSet (if ee then Map.keys res1 ++ Map.keys res2 else Map.keys res), not (Set.member (a,c) prev)])

obtainExtraQueue :: QueueAssignment -> [Int] -- list of data wire numbers
                    -> (Map.Map UntimedLPTerm Int,Map.Map UntimedLPTerm Int)
obtainExtraQueue (QueueAssignment qname _size inIrdy idata ouTrdy) vs
 = (enter , exit)
 where enter   = foldr combineAnd (translateTerm (T_AND inIrdy (T_QTRDY qname)))
                                  [translateTerm (idata !! v) | v<-vs] -- note: we could be much faster by not translating this as often as we do now
       exit    = foldr combineAnd (translateTerm (T_AND ouTrdy (T_QIRDY qname)))
                                  [translateTerm (T_QDATA qname v) | v<-vs]

obtainQueues :: Bool -> [QueueAssignment] -> [(QueueOrFlopNum, Map.Map UntimedLPTerm Int)]
obtainQueues _ [] = []
obtainQueues ee (o@(QueueAssignment nm _size inIrdy _idata ouTrdy) : lst)
 = if ee then (QN o Up, enter):(QN o Down,exit):obtainQueues ee lst
         else (QN o Val, enter .-. exit):obtainQueues ee lst
 where enter   = translateTerm (T_AND inIrdy (T_QTRDY nm))
       exit    = translateTerm (T_AND ouTrdy (T_QIRDY nm))

obtainFlops :: Bool -> [FlopAssignment] -> [(QueueOrFlopNum, Map.Map UntimedLPTerm Int)]
obtainFlops _ [] = []
obtainFlops tr (o@(FlopAssignment nm d):rst)
 = (if tr then [(FN o Up, posEdge),(FN o Down, negEdge)] else [(FN o Val, posEdge .-. negEdge .+. reset )]) ++
   (if null rst then [] else [(FFN o hr Up Up, toggle True True ),(FFN o hr Down Down, toggle False False )]  ++
     (if tr then [(FFN o hr Down Up, toggle False True ),(FFN o hr Up Down, toggle True False )] else [])) ++
   obtainFlops tr rst
 where reset = translateTerm (T_VALUE False)
       posEdge = translateTerm (T_AND (tnot (T_FLOPV nm)) d)
       negEdge = translateTerm (T_AND (T_FLOPV nm) (tnot d))
       toggle a b = translateTerm (T_AND (T_AND (a' (tnot (T_FLOPV nm))) (b' (tnot (T_FLOPV nm'))))
                                         (T_AND (a' d) (b' d')))
          where a' = (if a then id else tnot)
                b' = (if b then id else tnot)
       ~(FlopAssignment nm' d') = hr
       hr = head rst
        
data ValUpDown = Val | Up | Down deriving Show
data QueueOrFlopNum
  = QN QueueAssignment ValUpDown -- queue; Val=num, Up=enter or Down=exit
  | QNB QueueAssignment IntSet.IntSet ValUpDown -- queue with some bits set; Val=num, Up=enter or Down=exit
  | FN FlopAssignment ValUpDown -- flops
  | FFN FlopAssignment FlopAssignment ValUpDown ValUpDown -- two adjacent flops
  | N -- the `true' literal
  
dispQOFN :: QueueOrFlopNum -> [Char]
dispQOFN (QN  q   vup)  = ("Queue-"++name q++show vup)
dispQOFN (QNB q b vup)  = ("Queue-"++name q++show vup++show b)
dispQOFN (FN  f   vup)  = ("Flop-" ++name f++show vup)
dispQOFN (FFN f1 f2 v1 v2)=("Flop-" ++name f1++"-and-"++name f2++show v1++show v2)
dispQOFN N = "N"
