{-# OPTIONS_GHC -Wall -XFlexibleInstances #-}
module ToSMT (showSMT)  where
import ToINV (translateTerm, (.-.), (.+.), LPMap, UntimedLPTerm, getPrimitives, echelonReduce)
import Terms (Term(..), elimTopTerm)
import qualified Data.Set as Set
import qualified Data.Map.Lazy as Map
-- import Control.Applicative ()
-- import Data.Functor ()
import Data.Traversable (traverse,sequenceA)
import EParser (QueueAssignment(..),FlopAssignment(..),InputAssignment(..),OutputAssignment(..)
               ,OutputWire(..))
-- import RewriteStructure (rewrite,rewrite')
import Debug.Trace (trace)
import Prelude hiding (init,mapM)
import Formula((//),Formula(..),tosmt,getCoreList)
import Control.Monad.State (State(),get,put,runState)

type BoolVar = String
data PrimStmt = Perm Bool Term    -- Perm True x <=> x is permanently True; Perm False x <=> x is permanently False
              | PrimBoolVar BoolVar -- a boolean variable
              | IsConst Int StateVar -- states that this state variable is equal to this integer
              | InRange Int (LPMap StateVar) -- states that this sum of state variable is between 0 and this integer
data StateVar = QueueState String | FlopState String -- in the final LP, these will be integers indicating the current state. They are always assumed to be bounded!
                deriving (Ord, Eq, Show)

lookupOrInsert :: (Ord a) => a -> v -> Map.Map a v -> (v, Map.Map a v)
lookupOrInsert kx x t
  = case Map.insertLookupWithKey (\_ _ a -> a) kx x t of
      (Nothing,t') -> (x,t')
      (Just a, t') -> (a,t')

permNameState :: Bool -> Term -> State (Map.Map Term BoolVar) String
permNameState b t = do mp <- get
                       let (v,mp') = lookupOrInsert t (show (Map.size mp)) mp
                       put mp'
                       return (permName b v)

permName :: Bool -> String -> String
permName b v = "perm"++show b++"_"++v

primStmtToString :: PrimStmt -> State (Map.Map Term BoolVar) String
primStmtToString (Perm b t)       = permNameState b t
primStmtToString (PrimBoolVar bv) = return bv
primStmtToString (InRange 0 mp)   = return ("(= 0 "++showLPMap True getVarName mp++")")
primStmtToString (InRange i mp)   = return ("(and (<= 0 "++showLPMap True getVarName mp++") (>= "++show i++" "++showLPMap True getVarName mp++"))")
primStmtToString (IsConst i sv)   = return ("(= "++show i++" "++getVarName sv++")")

showSMT :: ([QueueAssignment],[FlopAssignment],[InputAssignment],[OutputAssignment],[OutputWire]) -> IO ()
showSMT x
 = putStrLn (concat ["(declare-fun " ++ s ++ " () "++showSMTType t++")\n" | (s,t) <- lst] ++
             "(assert " ++ 
             tosmt 8 id prb
             ++ ")\n(check-sat)\n(get-model)\n"
             ++ "(exit)"
            )
 where (prb,lst) = genProblem x
       showSMTType SMTInt  = "Int"
       showSMTType SMTBool = "Bool"
       showSMTType SMTReal = "Real"

getVarName :: StateVar -> String
getVarName (QueueState s) = "QState_" ++ s
getVarName (FlopState  s) = "FState_" ++ s

showLPMap :: (Num a, Ord a, Show a) => Bool -> (k -> String) -> Map.Map k a -> String
showLPMap isInt nm sv
 = case Map.toList sv of
     [] -> if isInt then "0" else "0.0"
     [(v,i)] -> showV i v
     lst -> "(+" ++ concat [' ':showV i v | (v,i)<-lst] ++ " )"
 where
  showV 1 v = nm v
  showV (-1) v = "(~ " ++ nm v ++ ")"
  showV i v = if i > 0 then "(* " ++ show i ++ " " ++ nm v ++ ")"
              else "(* (~ " ++ show (-i) ++ ") " ++ nm v ++ ")"

-- translate to formula: translation to the current value
translateToFormula :: (BoolVar -> Formula a) -> Term -> Formula a
translateToFormula _ (T_VALUE b  )   = if b then AND [] else OR []
translateToFormula _  T_RESET        = OR []
translateToFormula f (T_NOT a)       = NOT$ translateToFormula f a
translateToFormula f (T_AND a b)     = binAnd (translateToFormula f a) (translateToFormula f b)
translateToFormula f (T_OR  a b)     = binOr  (translateToFormula f a) (translateToFormula f b)
translateToFormula f (T_XOR a b)     = EQUIV  (translateToFormula f (tnot a)) (translateToFormula f b)
translateToFormula f o@(T_ITE _ _ _) = translateToFormula f (elimTopTerm o)
translateToFormula f x               = f (translateSoloTerm x)

translateSoloTerm :: Term -> String
translateSoloTerm x
 = case Map.keys (translateTerm x) of
           [a] -> (case (Set.toList a) of
                    [s] -> s
                    _ -> er)
           _ -> er
 where er = error ("Translation of "++show x++" resulted in too many terms for translateSoloTerm to handle (should result in one term).")

ttf :: Term -> Formula PrimStmt
ttf = translateToFormula (BASIC . PrimBoolVar)

tnot :: Term -> Term
tnot (T_NOT a) = a
tnot a = T_NOT a

binAnd,binOr :: Formula a -> Formula a -> Formula a
binAnd (AND lst1) (AND lst2) = AND (lst1 ++ lst2)
binAnd x (AND lst2) = AND (x:lst2)
binAnd (AND lst1) x = AND (lst1 ++ [x])
binAnd x y = AND [x,y]
binOr  (OR  lst1) (OR  lst2) = OR  (lst1 ++ lst2)
binOr  x (OR  lst2) = OR  (x:lst2)
binOr  (OR  lst1) x = OR  (lst1 ++ [x])
binOr  x y = OR  [x,y]

obtainFlops :: [FlopAssignment] -> ([(StateVar, LPMap UntimedLPTerm)], [Formula PrimStmt], [(BoolVar,Formula PrimStmt)])
obtainFlops [] = ([],[],[])
obtainFlops (FlopAssignment nm d:rst)
 = ( (FlopState  nm,(posEdge .-. negEdge) .+. reset):rst1
   , (AND [ EQUIV (BASIC (IsConst 1 (FlopState nm))) (ttf flopv)
                       , EQUIV (BASIC (IsConst 0 (FlopState nm))) (NOT (ttf flopv))
                       ] // "Setting the flop state of flop " ++ nm):rst2
   , (translateSoloTerm flopv, d'):rst3
   )
 where reset = translateTerm (T_VALUE False)
       posEdge = translateTerm (T_AND (T_NOT flopv) d)
       negEdge = translateTerm (T_AND flopv (T_NOT d))
       flopv = (T_FLOPV nm)
       d' = ttf d
       (rst1,rst2,rst3) = obtainFlops rst
       
-- invariant which is too weak: (= 0 (+ (~ QState_B1) QState_B2 (~ QState_B3) (~ QState_B4) (~ QState_B5) QState_B6 (~ QState_Buffer) ))
obtainQueues :: [QueueAssignment] -> ([(StateVar, LPMap UntimedLPTerm)], [Formula PrimStmt], [(BoolVar,Formula PrimStmt)])
obtainQueues qas = obtainForSets (Set.empty) [[]]
 where
   obtainForSets :: Set.Set [BoolVar] -> [[BoolVar]] -> ([(StateVar, LPMap UntimedLPTerm)], [Formula PrimStmt], [(BoolVar,Formula PrimStmt)])
   obtainForSets _ [] = ([],[],[])
   obtainForSets done (todo:lst)
    = if (Set.member todo done) then obtainForSets done lst else
       (res1 ++ ofs1, res2 ++ ofs2, res3 ++ ofs3)
    where
     (res1,res2,res3)
      = s3 [obtainQueue (concat v) qa | qa@(QueueAssignment q _ _ _ _) <- qas, Just v <- [traverse (getJustFor q) todo]] -- TODO: can be sped up by doing the second part just once. In bufferedVCRepeatable2 this just happens 3 times per queue.
     (ofs1,ofs2,ofs3)
      = obtainForSets (Set.insert todo done) (lst++ [[v | v <- Set.toList vs, take 5 v == "QDATA"] | vs <- Map.keys (Map.unions (map snd res1))])
   getJustFor q s = if take lnth s == ini then Just [read (drop lnth s)] else Nothing
    where ini  = "QDATA"++q++"-"
          lnth = length ini

s3 :: [([a],[b],[c])] -> ([a],[b],[c])
s3 [] = ([],[],[])
s3 ((as,bs,cs):rst) = (as++rst1,bs++rst2,cs++rst3) where (rst1,rst2,rst3) = s3 rst

obtainQueue :: [Int] -> QueueAssignment -> ([(StateVar, LPMap UntimedLPTerm)], [Formula PrimStmt], [(BoolVar,Formula PrimStmt)])
obtainQueue datas (QueueAssignment nm size inIrdy' idata ouTrdy)
 = ( [(stateVar,translateTerm enter .-. translateTerm exit)]
   , [ (AND$ (if null datas then
            [ EQUIV (BASIC (IsConst size stateVar)) (NOT (ttf inTrdy)) -- required for bufferedVCRepeatable2
            , EQUIV (BASIC (IsConst 0    stateVar)) (NOT (ttf ouIrdy)) -- required for bufferedVCRepeatable2
            , BASIC (InRange size (Map.singleton stateVar 1)) -- required for bufferedVCRepeatable2
            ] else
            [ IMPL  (BASIC (InRange 0 (Map.fromList [(stateVar,1),(total,-1)])))
                    (binOr (BASIC$ IsConst 0 total) (ttf ouType)) -- required for bufferedVCRepeatable2
            , IMPL  (binAnd (BASIC (IsConst 0 stateVar)) (ttf ouIrdy))
                    (NOT (ttf ouType)) -- required for bufferedVCRepeatable2
            , IMPL  (binAnd (NOT (BASIC$Perm False (T_AND inIrdy' inTrdy)) // "Something enters")
                            (BASIC (Perm False enter) // "But that something is never of this type")
                    )
                    ((BASIC$ IsConst 0 stateVar) // "Then none of the packets are of this type") -- required for bufferedVCRepeatable2
            , IMPL  (binAnd (NOT (BASIC (Perm False (T_AND inIrdy' inTrdy))) // "Something enters")
                            (BASIC (Perm False (T_XOR enter (T_AND inIrdy' inTrdy))) // "And that something is always of this type")
                    )
                    ((BASIC$ InRange 0 (Map.fromList [(stateVar,-1),(total,1)])) // "Then all of the packets are of this type")
                    -- required for bufferedVCRepeatable2
            ]) ++
            (IMPL (BASIC$Perm False enter)
                  (AND ((if null datas then
                          [ IMPL      (ttf inTrdy)  (BASIC$Perm True  inTrdy) -- required for bufferedVCRepeatable2
                          -- , IMPL (NOT (ttf inTrdy)) (BASIC$Perm False inTrdy) -- not required yet..
                          ] else []) ++
                          [IMPL      (ttf ouIrdy)  (BASIC$Perm True  ouIrdy) -- required for bufferedVCRepeatable2
                          --, IMPL (NOT (ttf ouIrdy)) (BASIC$Perm False ouIrdy) -- not required yet..
                          ]))
             // "Queue does not let anything through") :
            (IMPL (BASIC$Perm True inIrdy)
                  (if null datas then (BASIC$Perm True ouIrdy)
                   else (binOr (BASIC$Perm False ouTrdy) (BASIC$Perm True ouIrdy))
                  ) // "Queue source always has packets, so its output always has packets too") : -- required for bufferedVCRepeatable2
            (if null datas then -- the expression is still true in case of datas, but it will result in the same expression every time
              [ IMPL (BASIC$Perm True ouTrdy)
                     (BASIC$Perm True inTrdy) // "Queue target is always being accepted"
              ] else [] -- BASIC$ InRange size (Map.fromList [(stateVar,-1),(total,1)]) -- not required yet
              )
         ) // "Setting the queue state of " ++ nm ]
   , (if null datas then
        [ (translateSoloTerm inTrdy, OR [ ttf ouTrdy
                                        , AND [ttf inTrdy, NOT (ttf inIrdy)]
                                        , AND [ttf inTrdy, NOT (BASIC$ IsConst (size-1) stateVar)]
                                        ])
        , (translateSoloTerm ouIrdy, OR [ ttf inIrdy
                                        , AND [ttf ouIrdy, NOT (ttf ouTrdy)]
                                        , AND [ttf ouIrdy, NOT (BASIC$ IsConst 1 stateVar)]
                                        ])
        ]
      else []
     )
   )
 where enter    = T_AND inIrdy inTrdy                                      -- packet of specific type enters
       exit     = T_AND ouTrdy ouIrdy                                      -- packet of specific type leaves
       inIrdy   = foldr T_AND inIrdy' (map ((!!) idata) datas)             -- input of this specific type is ready
       inTrdy   = T_QTRDY nm                                       -- the queue accepts input
       ouIrdy   = T_AND (T_QIRDY nm) ouType -- output of specific type is ready
       ouType   = foldr T_AND (T_VALUE True) (map (\x -> T_QDATA nm x) datas)
       stateVar = QueueState (nm ++ if (null datas) then "" else (concat ["-"++show d|d<-datas]))
       total    = QueueState nm

splitAround :: [t] -> [t] -> [([t], t)]
splitAround pre (a:as) = (pre++as,a):splitAround (pre++[a]) as
splitAround _ [] = []

putPermInfoAndGetDecs :: [Formula PrimStmt]  -- formulas to be translated
                      -> [LPMap UntimedLPTerm] -- variables that are constant on average (for the lasso property)
                      -> Map.Map BoolVar (Formula PrimStmt) -- next state of these booleans
                      -> ([Formula String],
                          [(String,SMTType)] -- declarations for persistancy booleans, translateTerm-variables and next state booleans
                          )
putPermInfoAndGetDecs fs cs nx
 = (f ++ map fst varInfo ++
   [ AND$[ EQUIV (BASIC$ permName True nm) (BASIC$ "(= 1.0 "++ lpMap ++")")
         , EQUIV (BASIC$ permName False nm) (BASIC$ "(= 0.0 "++ lpMap ++")")
         -- , BASIC$ "(<= 0.0 "++lpMap++")" -- does not seem to be required.. is it derived or given elsewhere?
         -- , BASIC$ "(>= 1.0 "++lpMap++")"
         -- , IMPL (BASIC$ permName True nm) (translateToFormula BASIC t) -- follows from the primitives I guess?
         -- , IMPL (BASIC$ permName False nm) (NOT (translateToFormula BASIC t))
         ] ++
         (case ttn t of
            Just t' -> [ IMPL (BASIC$ permName True nm) t'
                       , IMPL (BASIC$ permName False nm) (NOT t')]
            Nothing -> [])
   | ((t,v), nm) <- perms, let lpMap = showLPMap False getBvSetName v]
   ++
   [ AND [ BASIC$ "(= 0.0 "++showLPMap False getBvSetName v++")"
         | v <- cs] // "lasso property"
   ] ++
   [ AND [ EQUIV (BASIC$ "next"++nm) v
         | (nm,v) <- Map.toList nextStates
         ] // "next states"
   ]
   , decs) -- these are just the variables created here
 where
   ttn :: Term -> Maybe (Formula String)
   ttn t = tTmp2
    where
     tTmp :: Formula (Maybe String)
     tTmp = translateToFormula (BASIC . lookupNext) t
     tTmp2 :: Maybe (Formula String)
     tTmp2 = traverse id tTmp
   lookupNext :: BoolVar -> (Maybe String)
   lookupNext x
     = do _ <- Map.lookup x nextStates
          return$ "next"++x
   
   (nextStates, prmap) = runState (sequenceA (fmap (traverse primStmtToString) nx)) Map.empty
   (f, tmap) = runState (sequenceA (fmap (traverse primStmtToString) fs)) prmap
   
   decs = [(permName b nm, SMTBool) | (_,nm)<-perms, b<-[False, True]] ++
          [(v, SMTReal) | (_,v) <- varInfo] ++
          [("next"++v, SMTBool) | v <- Map.keys nx]
   perms = [((t,translateTerm t), nm) | (t,nm) <- Map.toList tmap]
   
   allBVSets = Map.keysSet (Map.unions ([tt | ((_,tt),_)<-perms] ++ cs))
   allBVSetList :: [(Set.Set BoolVar, Int)] -- give a type to [0..], it is only used in "show"
   allBVSetList = zip (Set.toList allBVSets) [0..]
   bvSetNr = Map.fromAscList allBVSetList
   getBvSetName x = case Map.lookup x bvSetNr of
                      Nothing -> error ("The set "++show x++" is not found or named")
                      Just x' -> nameOf x'
   varInfo = map getVarInfo (splitAround [] allBVSetList)
   nameOf i' = "var"++show i'
   getVarInfo (other,(a,i))
    = (binAnd (case Set.toList a of
        [] -> BASIC$ "(= 1.0 "++nm++")" -- constant one
        [s] -> AND [ if isUnknown s then (AND [ NOT is0 -- fairness.. but the input may still be eager!
                                                           , IMPL is1 (BASIC s)
                                                           ])
                                                 else (AND [ IMPL is0 (NOT (BASIC s))
                                                           , IMPL is1 (BASIC s)
                                                           ])
                   , BASIC$ "(<= 0.0 "++nm++")"
                   , BASIC$ "(>= 1.0 "++nm++")"
                   ] // "permanence of "++s
        lst -> AND [ IMPL is0 (NOT $ AND (map BASIC lst))
                   , -- if (hasUnknowns lst) then NOT is1 else
                     IMPL is1 (AND (map BASIC lst))
                   , BASIC$ "(<= 0.0 "++nm++")"
                   -- , BASIC$ "(>= 1.0 "++nm++")" -- not needed, probably derived from its subsets
                   ] // "permanence of "++show lst)
              (AND [ (case (diff<b, Map.lookup diff bvSetNr) of
                       (True, Just diffNr)
                         -> binAnd (BASIC$ "(>= (+ 1.0 "++nm++") (+ "++nameOf bNr++" "++nameOf diffNr++"))") -- required
                       _ -> id)
                     (BASIC ("(<= "++nm++" "++(nameOf bNr)++")")) -- required
                   | (b,bNr) <- other -- happens 149^2 times for bufferedVCRepeatable2
                   , Set.isProperSubsetOf b a -- we could reduce the number of these comparisons by computing "other" in a smarter way
                   , let diff = Set.difference a b
                   ])
      , nm)
    where nm = nameOf i
          is0 = BASIC ("(= 0.0 "++nm++")")
          is1 = BASIC ("(= 1.0 "++nm++")")

isUnknown :: [Char] -> Bool
isUnknown s = take 5 s == "UNKNO"

data SMTType = SMTInt | SMTBool | SMTReal
genProblem :: ([QueueAssignment],[FlopAssignment],[InputAssignment],[OutputAssignment],[OutputWire]) -- the network
           -> (Formula String, [(String,SMTType)])
genProblem (queues,flops,_ins,_outs,_wires)
 = ( AND sformula, decs ++ [(getVarName sv, SMTInt) | (sv,_) <- svs] ++ uniqueBools [s | (PrimBoolVar s) <- getCoreList (AND$ formula ++ map snd nxtbv)]
   ) 
 where formula = ( (OR [ BASIC (Perm True dl) // "Deadness of queue " ++ nm
                       | (nm, dl) <- deadlocks ]) :
                   (AND invariants // "Invariants for this network") :
                   (AND fmlas // "All flops and queues") :
                   []
                 )
       (sformula, decs) = putPermInfoAndGetDecs formula stateExprs nextBoolVarValues
       uniqueBools lst = [(s,SMTBool) | s<-Set.toList (Set.fromList lst)]
       nextBoolVarValues = Map.fromListWith (error "Duplicate next bool values!") nxtbv
       deadlocks  = [(nm,T_AND (T_QIRDY nm) (T_NOT t)) | (QueueAssignment nm _ _ _ t) <- queues]
       stateExprs = map snd svs
       (svs,fmlas,nxtbv) = s3 [obtainQueues queues, obtainFlops flops]
       primitives = Set.toList (getPrimitives (map snd svs))
       nrOfPrims  = length primitives
       primNr     = Map.fromList (zip primitives [0..])
       lookupKey :: Int -> Int -> Maybe (StateVar, Int)
       lookupKey k i = if k < nrOfPrims then Nothing else Just (fst (svs !! (k - nrOfPrims)),i)
       invariants = [ BASIC (InRange 0 (Map.fromListWith (error "Duplicate state names!")$ Map.elems er))
                    | (Just er) <- 
                       map (Map.traverseWithKey lookupKey)
                           ({-# SCC "echelonReduce" #-}
                           echelonReduce [ Map.fromList ([((Map.!) primNr p,v) | (p,v)<-Map.toList qi, v/=0] ++ [(j,1)])
                                          | (j,(_,qi))<- zip [nrOfPrims..] svs])
                    ]
